import json
from pathlib import Path
from typing import Any

import regex

from dsp_tools.commands.excel2json.json_header import get_json_header
from dsp_tools.commands.excel2json.lists import excel2lists
from dsp_tools.commands.excel2json.new_lists.make_new_lists import new_excel2lists
from dsp_tools.commands.excel2json.properties import excel2properties
from dsp_tools.commands.excel2json.resources import excel2resources
from dsp_tools.models.exceptions import UserError


def excel2json(
    data_model_files: str,
    path_to_output_file: str,
) -> bool:
    """
    Converts a folder containing Excel files into a JSON data model file. The folder must be structured like this:

    ::

        data_model_files
        ├── json_header.xlsx (optional)
        ├── lists
        │   ├── de.xlsx
        │   └── en.xlsx
        └── onto_name (onto_label)
            ├── properties.xlsx
            └── resources.xlsx

    The names of the files must be exactly like in the example. The folder "lists" can be missing, because it is
    optional to have lists in a DSP project. Only XLSX files are allowed.

    Args:
        data_model_files: path to the folder (called "data_model_files" in the example)
        path_to_output_file: path to the file where the output JSON file will be saved

    Raises:
        UserError: if something went wrong
        BaseError: if something went wrong

    Returns:
        True if everything went well
    """

    listfolder, onto_folders = _validate_folder_structure_get_filenames(data_model_files)

    overall_success, project = _create_project_json(data_model_files, listfolder, onto_folders)

    with open(path_to_output_file, "w", encoding="utf-8") as f:
        json.dump(project, f, indent=4, ensure_ascii=False)

    print(f"JSON project file successfully saved at {path_to_output_file}")

    return overall_success


def _validate_folder_structure_get_filenames(data_model_files: str) -> tuple[list[Path], list[Path]]:
    if not Path(data_model_files).is_dir():
        raise UserError(f"ERROR: {data_model_files} is not a directory.")
    sub_folders = [x for x in Path(data_model_files).glob("*") if _non_hidden(x) and x.is_dir()]
    files_to_process = []
    onto_folders, onto_files_to_process = _get_and_validate_onto_folder(Path(data_model_files), sub_folders)
    files_to_process.extend(onto_files_to_process)
    listfolder, lists_to_process = _get_validate_list_folder(data_model_files, sub_folders)
    files_to_process.extend(lists_to_process)
    if len(onto_folders) + len(listfolder) != len(sub_folders):
        raise UserError(
            f"The only allowed subfolders in '{data_model_files}' are 'lists' "
            "and folders that match the pattern 'onto_name (onto_label)'"
        )
    if (json_header := Path(data_model_files) / Path("json_header.xlsx")).exists():
        files_to_process.append(str(json_header))
    print("The following files will be processed:")
    print(*(f" - {file}" for file in files_to_process), sep="\n")
    return listfolder, onto_folders


def _get_validate_list_folder(data_model_files: str, folder: list[Path]) -> tuple[list[Path], list[str]]:
    processed_files: list[str] = []
    listfolder = [x for x in folder if x.is_dir() and x.name == "lists"]
    if listfolder:
        listfolder_contents = [x for x in Path(listfolder[0]).glob("*") if _non_hidden(x)]
        if not all(regex.search(r"(de|en|fr|it|rm).xlsx", file.name) for file in listfolder_contents):
            raise UserError(
                f"The only files allowed in '{data_model_files}/lists' are en.xlsx, de.xlsx, fr.xlsx, it.xlsx, rm.xlsx"
            )
        processed_files = [f"{data_model_files}/lists/{file.name}" for file in listfolder_contents]
    return listfolder, processed_files


def new_excel2json(
    data_model_files: str,
    path_to_output_file: str,
) -> bool:
    """
    Converts a folder containing Excel files into a JSON data model file. The folder must be structured like this:

    ::

        data_model_files
        ├── json_header.xlsx (optional)
        ├── lists
        │   ├── list.xlsx
        │   └── list_1.xlsx
        └── onto_name (onto_label)
            ├── properties.xlsx
            └── resources.xlsx

    The names of the files must be exactly like in the example. The folder "lists" can be missing, because it is
    optional to have lists in a DSP project. Only XLSX files are allowed.

    Args:
        data_model_files: path to the folder (called "data_model_files" in the example)
        path_to_output_file: path to the file where the output JSON file will be saved

    Raises:
        InputError: if something went wrong

    Returns:
        True if everything went well
    """

    listfolder, onto_folders = _new_validate_folder_structure_and_get_filenames(Path(data_model_files))

    overall_success, project = _new_create_project_json(data_model_files, onto_folders, listfolder)

    with open(path_to_output_file, "w", encoding="utf-8") as f:
        json.dump(project, f, indent=4, ensure_ascii=False)

    print(f"JSON project file successfully saved at {path_to_output_file}")

    return overall_success


def _new_validate_folder_structure_and_get_filenames(data_model_files: Path) -> tuple[Path | None, list[Path]]:
    if not data_model_files.is_dir():
        raise UserError(f"ERROR: {data_model_files} is not a directory.")
    folder = [x for x in data_model_files.glob("*") if _non_hidden(x)]
    processed_files = []
    onto_folders, processed_onto = _get_and_validate_onto_folder(data_model_files, folder)
    processed_files.extend(processed_onto)
    listfolder, processed_lists = _new_get_and_validate_list_folder(data_model_files)
    processed_files.extend(processed_lists)
    print("The following files will be processed:")
    print(*(f" - {file}" for file in processed_files), sep="\n")
    return listfolder, onto_folders


def _get_and_validate_onto_folder(data_model_files: Path, folder: list[Path]) -> tuple[list[Path], list[str]]:
    processed_files = []
    onto_folders = [x for x in folder if x.is_dir() and regex.search(r"([\w.-]+) \(([\w.\- ]+)\)", x.name)]
    if not onto_folders:
        raise UserError(
            f"'{data_model_files}' must contain at least one subfolder named after the pattern 'onto_name (onto_label)'"
        )
    for onto_folder in onto_folders:
        contents = sorted([x.name for x in Path(onto_folder).glob("*") if _non_hidden(x)])
        if contents != ["properties.xlsx", "resources.xlsx"]:
            raise UserError(
                f"ERROR: '{data_model_files}/{onto_folder.name}' must contain one file 'properties.xlsx' "
                "and one file 'resources.xlsx', but nothing else."
            )
        processed_files.extend([f"{data_model_files}/{onto_folder.name}/{file}" for file in contents])
    return onto_folders, processed_files


def _new_get_and_validate_list_folder(
    data_model_files: Path,
) -> tuple[Path | None, list[str]]:
    if not (list_dir := (data_model_files / "lists")).is_dir():
        return None, []
    processed_files = [str(file) for file in list_dir.glob("*list*.xlsx") if _non_hidden(file)]
    return list_dir, processed_files


def _non_hidden(path: Path) -> bool:
    return not regex.search(r"^(\.|~\$).+", path.name)


def _create_project_json(
    data_model_files: str, listfolder: list[Path], onto_folders: list[Path]
) -> tuple[bool, dict[str, Any]]:
    overall_success = True
    lists, success = excel2lists(excelfolder=f"{data_model_files}/lists") if listfolder else (None, True)
    if not success:
        overall_success = False
    ontologies, success = _get_ontologies(data_model_files, onto_folders)
    if not success:
        overall_success = False
    project = get_json_header(Path(data_model_files) / "json_header.xlsx").to_dict()
    if lists:
        project["project"]["lists"] = lists
    project["project"]["ontologies"] = ontologies
    return overall_success, project


def _new_create_project_json(
    data_model_files: str, onto_folders: list[Path], list_folder: Path | None
) -> tuple[bool, dict[str, Any]]:
    overall_success = True
    lists = None
    if list_folder:
        lists, success = new_excel2lists(list_folder)
        if not success:
            overall_success = False
    ontologies, success = _get_ontologies(data_model_files, onto_folders)
    if not success:
        overall_success = False
    project = get_json_header(Path(data_model_files) / "json_header.xlsx").to_dict()
    if lists:
        project["project"]["lists"] = lists
    project["project"]["ontologies"] = ontologies
    return overall_success, project


def _get_ontologies(data_model_files: str, onto_folders: list[Path]) -> tuple[list[dict[str, Any]], bool]:
    success = True
    ontologies = []
    for onto_folder in onto_folders:
        name, label = regex.search(r"([\w.-]+) \(([\w.\- ]+)\)", onto_folder.name).groups()  # type: ignore[union-attr]
        resources, success1 = excel2resources(f"{data_model_files}/{onto_folder.name}/resources.xlsx")
        properties, success2 = excel2properties(f"{data_model_files}/{onto_folder.name}/properties.xlsx")
        if not success1 or not success2:
            success = False
        ontologies.append(
            {
                "name": name,
                "label": label,
                "properties": properties,
                "resources": resources,
            }
        )
    return ontologies, success



import importlib.resources
import shutil
from pathlib import Path

from dsp_tools.models.exceptions import UserError


def generate_template_repo() -> bool:
    """
    Copies the folder src/dsp_tools/resources/0100-template-repo to the working directory.

    Raises:
        UserError: If a folder with the same name already exists in the working directory

    Returns:
        True if the folder could be copied without problems
    """
    # copy contents of src/dsp_tools/resources/0100-template-repo to cwd/template-repo
    template_path_of_user = Path("0100-template-repo")
    try:
        template_path_of_user.mkdir()
    except OSError:
        raise UserError("A folder '0100-template-repo' already exists in your current working directory.") from None
    template_path_of_distribution = importlib.resources.files("dsp_tools").joinpath("resources/0100-template-repo")
    for file in template_path_of_distribution.iterdir():
        with importlib.resources.as_file(file) as f:
            shutil.copy(f, template_path_of_user / file.name)
        print(f"Created {template_path_of_user / file.name}")

    return True



from __future__ import annotations

from typing import cast
from uuid import uuid4

from lxml import etree

from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLProperty
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import create_info_from_xml_for_graph
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import generate_upload_order
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import make_graph
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStash
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStash
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStashItem
from dsp_tools.commands.xmlupload.stash.stash_models import Stash


def _stash_standoff(
    res_id: str,
    restype: str,
    link_prop: XMLProperty,
    stash_lookup: dict[str, list[str]],
) -> list[StandoffStashItem]:
    stashed_items = []
    for value in link_prop.values:
        if value.link_uuid not in stash_lookup[res_id]:
            continue
        # value.value is a KnoraStandoffXml text with problematic links.
        # stash it, then replace the problematic text with a UUID
        standoff_xml = cast(FormattedTextValue, value.value)
        uuid = str(uuid4())
        standoff_stash_item = StandoffStashItem(
            res_id=res_id,
            res_type=restype,
            uuid=uuid,
            prop_name=link_prop.name,
            value=standoff_xml,
        )
        value.value = FormattedTextValue(uuid)
        stashed_items.append(standoff_stash_item)
    return stashed_items


def _stash_resptr(
    res_id: str,
    restype: str,
    link_prop: XMLProperty,
    stash_lookup: dict[str, list[str]],
    permission_lookup: dict[str, Permissions],
) -> list[LinkValueStashItem]:
    stashed_items = []
    for value in link_prop.values.copy():
        if value.link_uuid not in stash_lookup[res_id]:
            continue
        permission = str(permission_lookup[value.permissions]) if value.permissions else None
        # value.value is the ID of the target resource. stash it, then delete it
        link_stash_item = LinkValueStashItem(
            res_id=res_id,
            res_type=restype,
            prop_name=link_prop.name,
            target_id=str(value.value),
            permission=permission,
        )
        link_prop.values.remove(value)
        stashed_items.append(link_stash_item)
    return stashed_items


def stash_circular_references(
    resources: list[XMLResource],
    stash_lookup: dict[str, list[str]],
    permission_lookup: dict[str, Permissions],
) -> Stash | None:
    """
    Stashes problematic resource-references from a list of resources.
    The resources are modified in-place.

    Args:
        resources: all resources of the XML file
        stash_lookup: A dictionary which maps the resources that have stashes to the UUIDs of the stashed links
        permission_lookup: A dictionary which maps the permissions of the stashed links to their string representation

    Returns:
        stash: an object that contains the stashed references

    Raises:
        ValueError: If a link property of one of the resources is not "text" or "resptr"
    """
    stashed_standoff_values: list[StandoffStashItem] = []
    stashed_link_values: list[LinkValueStashItem] = []

    for res in resources:
        if res.res_id not in stash_lookup:
            continue
        for link_prop in res.get_props_with_links():
            if link_prop.valtype == "text":
                standoff_stash_item = _stash_standoff(res.res_id, res.restype, link_prop, stash_lookup)
                stashed_standoff_values.extend(standoff_stash_item)
            elif link_prop.valtype == "resptr":
                link_stash_item = _stash_resptr(res.res_id, res.restype, link_prop, stash_lookup, permission_lookup)
                stashed_link_values.extend(link_stash_item)
            else:
                raise ValueError(f"Unknown value type: '{link_prop.valtype}' (should be 'text' or 'resptr')")

            if len(link_prop.values) == 0:
                # if all values of a link property have been stashed, the property needs to be removed
                res.properties.remove(link_prop)

    standoff_stash = StandoffStash.make(stashed_standoff_values)
    link_value_stash = LinkValueStash.make(stashed_link_values)
    return Stash.make(standoff_stash, link_value_stash)


def identify_circular_references(root: etree._Element) -> tuple[dict[str, list[str]], list[str]]:
    """
    Identifies problematic resource-references inside an XML tree.
    A reference is problematic if it creates a circle (circular references).
    The XML tree is modified in-place:
    A reference UUID is added to each XML element that contains a link (`<resptr>` or `<text>`).

    Args:
        root: the root element of the parsed XML document

    Returns:
        stash_lookup: A dictionary which maps the resources that have stashes to the UUIDs of the stashed links
        upload_order: A list of resource IDs in the order in which they should be uploaded
    """
    resptr_links, xml_links, all_resource_ids = create_info_from_xml_for_graph(root)
    graph, node_to_id, edges = make_graph(resptr_links, xml_links, all_resource_ids)
    stash_lookup, upload_order, _ = generate_upload_order(graph, node_to_id, edges)
    return stash_lookup, upload_order



# mypy: disable-error-code="method-assign"

import json
from dataclasses import dataclass
from importlib.metadata import version
from typing import Any
from typing import Callable
from typing import cast
from unittest.mock import Mock
from unittest.mock import patch

import pytest
import regex
from requests import ReadTimeout
from requests import RequestException

from dsp_tools.models.exceptions import PermanentConnectionError
from dsp_tools.models.exceptions import PermanentTimeOutError
from dsp_tools.models.exceptions import UserError
from dsp_tools.utils.connection_live import ConnectionLive
from dsp_tools.utils.connection_live import RequestParameters


@dataclass
class SessionMock:
    responses: tuple[Any, ...]
    counter = 0

    def request(self, **kwargs: Any) -> Any:  # noqa: ARG002
        response = self.responses[self.counter]
        self.counter += 1
        if isinstance(response, BaseException):
            raise response
        return response


@dataclass
class ResponseMock:
    status_code: int
    headers: dict[str, Any]
    text: str


def test_log_in_log_out() -> None:
    con = ConnectionLive("http://example.com/")
    con.post = Mock(return_value={"token": "token"})
    con.login("root@example.com", "test")
    assert con.post.call_args.kwargs["route"] == "/v2/authentication"
    assert con.token == "token"
    assert con.session.headers["Authorization"] == "Bearer token"
    con.delete = Mock()
    con.logout()
    assert con.token is None
    assert "Authorization" not in con.session.headers


def test_log_in_bad_credentials() -> None:
    con = ConnectionLive("http://example.com/")
    con._log_response = Mock()
    con.session.request = Mock(return_value=Mock(status_code=401))
    with patch("dsp_tools.utils.connection_live.time.sleep") as sleep_mock:
        with pytest.raises(UserError, match=regex.escape("Username and/or password are not valid")):
            con.login("invalid@example.com", "wrong")
        sleep_mock.assert_not_called()


def test_post() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    con.post(route="/v2/resources")
    expected_params: RequestParameters = con._try_network_action.call_args.args[0]
    assert expected_params.method == "POST"
    assert expected_params.url == "http://example.com/v2/resources"
    assert expected_params.data is None
    assert expected_params.headers is None
    assert expected_params.files is None


def test_post_with_data() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    con.post(route="/v2/resources", data={"foo": "bar"})
    expected_params: RequestParameters = con._try_network_action.call_args.args[0]
    assert expected_params.method == "POST"
    assert expected_params.url == "http://example.com/v2/resources"
    assert expected_params.data == {"foo": "bar"}
    assert expected_params.headers == {"Content-Type": "application/json; charset=UTF-8"}
    assert expected_params.files is None


def test_post_with_file() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    file = {"file": ("path/to/file.jpg", b"some bytes")}
    con.post(route="/upload", files=file)
    expected_params: RequestParameters = con._try_network_action.call_args.args[0]
    assert expected_params.method == "POST"
    assert expected_params.url == "http://example.com/upload"
    assert expected_params.data is None
    assert expected_params.headers is None
    assert expected_params.files == file


def test_get() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    con.get(route="/admin/groups")
    expected_params: RequestParameters = con._try_network_action.call_args.args[0]
    assert expected_params.method == "GET"
    assert expected_params.url == "http://example.com/admin/groups"
    assert expected_params.data is None
    assert expected_params.headers is None
    assert expected_params.files is None


def test_put() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    con.put(route="/v2/values")
    expected_params: RequestParameters = con._try_network_action.call_args.args[0]
    assert expected_params.method == "PUT"
    assert expected_params.url == "http://example.com/v2/values"
    assert expected_params.data is None
    assert expected_params.headers is None
    assert expected_params.files is None


def test_put_with_data() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    con.put(route="/v2/values", data={"foo": "bar"})
    expected_params: RequestParameters = con._try_network_action.call_args.args[0]
    assert expected_params.method == "PUT"
    assert expected_params.url == "http://example.com/v2/values"
    assert expected_params.data == {"foo": "bar"}
    assert expected_params.headers == {"Content-Type": "application/json; charset=UTF-8"}
    assert expected_params.files is None


def test_delete() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    con.delete(route="/v2/values")
    expected_params: RequestParameters = con._try_network_action.call_args.args[0]
    assert expected_params.method == "DELETE"
    assert expected_params.url == "http://example.com/v2/values"
    assert expected_params.data is None
    assert expected_params.headers is None
    assert expected_params.files is None


def test_default_timeout() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    for method in (con.post, con.put):
        method = cast(Callable[..., Any], method)
        method(route="/v2/resources")
        expected_params: RequestParameters = con._try_network_action.call_args.args[0]
        assert expected_params.timeout == con.timeout_put_post, f"Method '{method.__name__}' failed"
    for method in (con.get, con.delete):  # type: ignore[assignment]
        method = cast(Callable[..., Any], method)
        method(route="/v2/resources")
        expected_params = con._try_network_action.call_args.args[0]
        assert expected_params.timeout == con.timeout_get_delete, f"Method '{method.__name__}' failed"


def test_custom_timeout() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    con.post(route="/v2/resources", timeout=1)
    expected_params: RequestParameters = con._try_network_action.call_args.args[0]
    assert expected_params.timeout == 1


def test_custom_header() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    for method in (con.post, con.put, con.get, con.delete):
        method = cast(Callable[..., Any], method)
        method(route="/v2/resources", headers={"foo": "bar"})
        expected_params: RequestParameters = con._try_network_action.call_args.args[0]
        assert expected_params.headers == {"foo": "bar"}, f"Method '{method.__name__}' failed"


def test_custom_content_type() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    for method in (con.post, con.put, con.get, con.delete):
        method = cast(Callable[..., Any], method)
        method(route="/v2/resources", headers={"Content-Type": "bar"})
        expected_params: RequestParameters = con._try_network_action.call_args.args[0]
        assert expected_params.headers == {"Content-Type": "bar"}, f"Method '{method.__name__}' failed"


def test_server_without_trailing_slash() -> None:
    con = ConnectionLive("http://example.com")
    con._try_network_action = Mock()
    for method in (con.post, con.put, con.get, con.delete):
        method = cast(Callable[..., Any], method)
        method(route="/v2/resources")
        expected_params: RequestParameters = con._try_network_action.call_args.args[0]
        assert expected_params.url == "http://example.com/v2/resources", f"Method '{method.__name__}' failed"


def test_route_without_leading_slash() -> None:
    con = ConnectionLive("http://example.com/")
    con._try_network_action = Mock()
    for method in (con.post, con.put, con.get, con.delete):
        method = cast(Callable[..., Any], method)
        method(route="v2/resources")
        expected_params: RequestParameters = con._try_network_action.call_args.args[0]
        assert expected_params.url == "http://example.com/v2/resources", f"Method '{method.__name__}' failed"


def test_server_and_route_without_slash() -> None:
    con = ConnectionLive("http://example.com")
    con._try_network_action = Mock()
    for method in (con.post, con.put, con.get, con.delete):
        method = cast(Callable[..., Any], method)
        method(route="v2/resources")
        expected_params: RequestParameters = con._try_network_action.call_args.args[0]
        assert expected_params.url == "http://example.com/v2/resources", f"Method '{method.__name__}' failed"


def test_anonymize_different_keys() -> None:
    con = ConnectionLive("foo")
    assert con._anonymize(None) is None
    assert con._anonymize({"foo": "bar"}) == {"foo": "bar"}
    assert con._anonymize({"token": "uk7m20-8gqn8"}) == {"token": "uk7m2[+7]"}
    assert con._anonymize({"Set-Cookie": "uk7m20-8gqn8"}) == {"Set-Cookie": "uk7m2[+7]"}
    assert con._anonymize({"Authorization": "Bearer uk7m20-8gqn8"}) == {"Authorization": "Bearer uk7m2[+7]"}
    assert con._anonymize({"password": "uk7m20-8gqn8"}) == {"password": "************"}


def test_anonymize_doesnt_mutate_original() -> None:
    con = ConnectionLive("foo")
    random = {"foo": "bar"}
    assert con._anonymize(random) is not random


def test_anonymize_different_lengths() -> None:
    con = ConnectionLive("foo")
    assert con._anonymize({"token": "uk7m20-8gqn8ir7e30"}) == {"token": "uk7m2[+13]"}
    assert con._anonymize({"token": "uk7m2"}) == {"token": "*****"}
    assert con._anonymize({"token": "u"}) == {"token": "*"}


def test_try_network_action() -> None:
    con = ConnectionLive("http://example.com/")
    response_expected = Mock(status_code=200)
    con.session.request = Mock(return_value=response_expected)
    con._log_request = Mock()
    con._log_response = Mock()
    params = RequestParameters(method="GET", url="http://example.com/", timeout=1)
    response = con._try_network_action(params)
    assert response == response_expected
    con._log_request.assert_called_once_with(params)
    con.session.request.assert_called_once_with(**params.as_kwargs())
    con._log_response.assert_called_once_with(response_expected)


def test_try_network_action_timeout_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("DSP_TOOLS_TESTING", raising=False)  # in CI, this variable suppresses the retrying mechanism
    con = ConnectionLive("http://example.com/")
    responses = (TimeoutError(), ReadTimeout(), Mock(status_code=200))
    session_mock = SessionMock(responses)
    con.session = session_mock  # type: ignore[assignment]
    con._log_request = Mock()
    con._log_response = Mock()
    params = RequestParameters(method="GET", url="http://example.com/", timeout=1)
    expected_msg = regex.escape("A 'TimeoutError' occurred during the connection to the DSP server.")
    with pytest.raises(PermanentTimeOutError, match=expected_msg):
        con._try_network_action(params)


def test_try_network_action_connection_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("DSP_TOOLS_TESTING", raising=False)  # in CI, this variable suppresses the retrying mechanism
    con = ConnectionLive("http://example.com/")
    responses = (ConnectionError(), ConnectionError(), RequestException(), Mock(status_code=200))
    session_mock = SessionMock(responses)
    con.session = session_mock  # type: ignore[assignment]
    con._log_request = Mock()
    con._log_response = Mock()
    con._renew_session = Mock()
    params = RequestParameters(method="POST", url="http://example.com/", timeout=1)
    with patch("dsp_tools.utils.connection_live.time.sleep") as sleep_mock:
        response = con._try_network_action(params)
        assert [x.args[0] for x in sleep_mock.call_args_list] == [1, 2, 4]
    assert con._renew_session.call_count == len(session_mock.responses) - 1
    assert [x.args[0] for x in con._log_request.call_args_list] == [params] * len(session_mock.responses)
    con._log_response.assert_called_once_with(session_mock.responses[-1])
    assert response == session_mock.responses[-1]


def test_try_network_action_non_200(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("DSP_TOOLS_TESTING", raising=False)  # in CI, this variable suppresses the retrying mechanism
    con = ConnectionLive("http://example.com/")
    responses = (Mock(status_code=500, text=""), Mock(status_code=506, text=""), Mock(status_code=200, text=""))
    session_mock = SessionMock(responses)
    con.session = session_mock  # type: ignore[assignment]
    con._log_request = Mock()
    con._log_response = Mock()
    params = RequestParameters(method="POST", url="http://example.com/", timeout=1)
    with patch("dsp_tools.utils.connection_live.time.sleep") as sleep_mock:
        response = con._try_network_action(params)
        assert [x.args[0] for x in sleep_mock.call_args_list] == [1, 2]
    assert [x.args[0] for x in con._log_request.call_args_list] == [params] * len(session_mock.responses)
    assert [x.args[0] for x in con._log_response.call_args_list] == list(session_mock.responses)
    assert response == session_mock.responses[-1]


def test_try_network_action_in_testing_environment(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("DSP_TOOLS_TESTING", "true")  # automatically set in CI, but not locally
    con = ConnectionLive("http://example.com/")
    responses = (Mock(status_code=500, text=""), Mock(status_code=404, text=""), Mock(status_code=200, text=""))
    con.session = SessionMock(responses)  # type: ignore[assignment]
    con._log_request = Mock()
    con._log_response = Mock()
    params = RequestParameters(method="PUT", url="http://example.com/", timeout=1)
    with patch("dsp_tools.utils.connection_live.time.sleep") as sleep_mock:
        with pytest.raises(PermanentConnectionError):
            con._try_network_action(params)
        sleep_mock.assert_not_called()


def test_try_network_action_permanent_connection_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("DSP_TOOLS_TESTING", "true")  # automatically set in CI, but not locally
    con = ConnectionLive("http://example.com/")
    responses = (Mock(status_code=500, text=""),) * 7
    con.session = SessionMock(responses)  # type: ignore[assignment]
    con._log_request = Mock()
    con._log_response = Mock()
    params = RequestParameters(method="POST", url="http://example.com/", timeout=1)
    with pytest.raises(PermanentConnectionError):
        con._try_network_action(params)


def test_log_request() -> None:
    con = ConnectionLive("http://example.com/")
    con.session.headers = {"Authorization": "Bearer my-very-long-token"}
    params = RequestParameters(
        method="GET",
        url="http://example.com/",
        headers={"request-header": "request-value"},
        timeout=1,
        data={"password": "my-super-secret-password", "foo": "bar"},
    )
    expected_output = {
        "method": "GET",
        "url": "http://example.com/",
        "headers": {"Authorization": "Bearer my-ve[+13]", "request-header": "request-value"},
        "timeout": 1,
        "data": {"password": "************************", "foo": "bar"},
    }
    with patch("dsp_tools.utils.connection_live.logger.debug") as debug_mock:
        con._log_request(params)
        debug_mock.assert_called_once_with(f"REQUEST: {json.dumps(expected_output)}")


def test_log_response() -> None:
    con = ConnectionLive("http://example.com/")
    response_mock = ResponseMock(
        status_code=200,
        headers={"Set-Cookie": "KnoraAuthenticationMFYGSLT", "Content-Type": "application/json"},
        text=json.dumps({"token": "my-very-long-token", "foo": "bar"}),
    )
    expected_output = {
        "status_code": 200,
        "headers": {"Set-Cookie": "Knora[+21]", "Content-Type": "application/json"},
        "content": {"token": "my-ve[+13]", "foo": "bar"},
    }
    with patch("dsp_tools.utils.connection_live.logger.debug") as debug_mock:
        con._log_response(response_mock)  # type: ignore[arg-type]
        debug_mock.assert_called_once_with(f"RESPONSE: {json.dumps(expected_output)}")


def test_renew_session() -> None:
    con = ConnectionLive("http://example.com/", "my-super-secret-token")
    assert con.get_token() == "my-super-secret-token"
    assert con.session.headers["User-Agent"] == f'DSP-TOOLS/{version("dsp-tools")}'
    con._renew_session()
    assert con.get_token() == "my-super-secret-token"
    assert con.session.headers["User-Agent"] == f'DSP-TOOLS/{version("dsp-tools")}'


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from uuid import uuid4

from rdflib import RDF
from rdflib import RDFS
from rdflib import Graph
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

KNORA_API = Namespace("http://api.knora.org/ontology/knora-api/v2#")
API_SHAPES = Namespace("http://api.knora.org/ontology/knora-api/shapes/v2#")
DATA = Namespace("http://data/")


@dataclass
class DataRDF:
    triples: list[RDFTriples]

    def make_graph(self) -> Graph:
        g = Graph()
        for r in self.triples:
            g += r.make_graph()
        return g


class RDFTriples(ABC):
    def make_graph(self) -> Graph:
        raise NotImplementedError


@dataclass
class ResourceRDF(RDFTriples):
    res_iri: URIRef
    res_class: URIRef
    label: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        g.add((self.res_iri, RDF.type, self.res_class))
        g.add((self.res_iri, RDFS.label, self.label))
        return g


@dataclass
class ValueRDF(RDFTriples):
    prop_name: URIRef
    object_value: URIRef | Literal
    res_iri: URIRef

    def make_graph(self) -> Graph:
        raise NotImplementedError


@dataclass
class BooleanValueRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.BooleanValue))
        g.add((val_iri, KNORA_API.booleanValueAsBoolean, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class ColorValueRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.ColorValue))
        g.add((val_iri, KNORA_API.colorValueAsColor, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class DateValueRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.DateValue))
        g.add((val_iri, KNORA_API.valueAsString, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class DecimalValueRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.DecimalValue))
        g.add((val_iri, KNORA_API.decimalValueAsDecimal, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class GeonameValueRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.GeonameValue))
        g.add((val_iri, KNORA_API.geonameValueAsGeonameCode, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class IntValueRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.IntValue))
        g.add((val_iri, KNORA_API.intValueAsInt, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class LinkValueRDF(ValueRDF):
    object_value: URIRef

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.LinkValue))
        g.add((val_iri, API_SHAPES.linkValueHasTargetID, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class ListValueRDF(ValueRDF):
    prop_name: URIRef
    object_value: Literal
    list_name: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.ListValue))
        g.add((val_iri, API_SHAPES.listNodeAsString, self.object_value))
        g.add((val_iri, API_SHAPES.listNameAsString, self.list_name))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class SimpleTextRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.TextValue))
        g.add((val_iri, KNORA_API.valueAsString, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class RichtextRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.TextValue))
        g.add((val_iri, KNORA_API.textValueAsXml, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class TimeValueRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.TimeValue))
        g.add((val_iri, KNORA_API.timeValueAsTimeStamp, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g


@dataclass
class UriValueRDF(ValueRDF):
    object_value: Literal

    def make_graph(self) -> Graph:
        g = Graph()
        val_iri = DATA[str(uuid4())]
        g.add((val_iri, RDF.type, KNORA_API.UriValue))
        g.add((val_iri, KNORA_API.uriValueAsUri, self.object_value))
        g.add((self.res_iri, self.prop_name, val_iri))
        return g



import unittest

import jsonpath_ng
import jsonpath_ng.ext
import pytest
import regex
from pytest_unordered import unordered

from dsp_tools.commands.excel2json import resources as e2j
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.exceptions import InputError

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


excelfile = "testdata/excel2json/excel2json_files/test-name (test_label)/resources.xlsx"
output_from_method, _ = e2j.excel2resources(excelfile, None)


class TestExcelToResource(unittest.TestCase):
    def test_names(self) -> None:
        expected_names = [
            "Owner",
            "Title",
            "GenericAnthroponym",
            "FamilyMember",
            "MentionedPerson",
            "Alias",
            "Image",
            "Video",
            "Audio",
            "ZIP",
            "PDFDocument",
            "NoCardinalityClass",
        ]
        res_names = [match.value for match in jsonpath_ng.parse("$[*].name").find(output_from_method)]
        assert unordered(res_names) == expected_names

    def test_supers(self) -> None:
        expected_supers = [
            ["Resource", "dcterms:fantasy"],
            ["Resource"],
            ["Resource"],
            ["Resource"],
            ["Resource"],
            ["Resource"],
            ["StillImageRepresentation", "dcterms:image"],
            ["MovingImageRepresentation"],
            ["AudioRepresentation"],
            ["ArchiveRepresentation"],
            ["DocumentRepresentation"],
            ["Resource"],
        ]
        res_supers = [match.value for match in jsonpath_ng.parse("$[*].super").find(output_from_method)]
        assert unordered(res_supers) == expected_supers

    def test_labels(self) -> None:
        expected_labels = {
            "en": [
                "Owner",
                "Title",
                "Generic anthroponym",
                "Family member",
                "Mentioned person",
                "Alias",
                "Only English",
                "",
                "",
                "",
                "",
                "Class Without Cardinalities",
            ],
            "rm": [
                "Rumantsch",
                "Rumantsch",
                "Rumantsch",
                "Rumantsch",
                "Rumantsch",
                "Rumantsch",
                "",
                "",
                "",
                "",
                "Only Rumantsch",
                "",
            ],
        }
        res_labels_all = [match.value for match in jsonpath_ng.parse("$[*].labels").find(output_from_method)]
        res_labels = {lang: [label.get(lang, "").strip() for label in res_labels_all] for lang in ["en", "rm"]}
        self.assertDictEqual(res_labels, expected_labels)

    def test_image_labels(self) -> None:
        expected_labels_of_image = {"en": "Only English"}
        res_labels_of_image = jsonpath_ng.ext.parse('$[?name="Image"].labels').find(output_from_method)[0].value
        self.assertDictEqual(res_labels_of_image, expected_labels_of_image)

    def test_comments(self) -> None:
        expected_comments = {
            "comment_de": [
                "Ein seltsamer Zufall brachte mich in den Besitz dieses Tagebuchs.",
                "",
                "Only German",
                "",
                "",
                "",
                "Bild",
                "Video",
                "Audio",
                "ZIP",
                "PDF-Dokument",
                "",
            ],
            "comment_fr": [
                "Un étrange hasard m'a mis en possession de ce journal.",
                "",
                "",
                "Only French",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
            ],
        }
        # make sure the lists of the json comments contain a blank string,
        # even if there is no "comments" section at all in this resource
        res_comments = {
            f"comment_{lang}": [resource.get("comments", {}).get(lang, "").strip() for resource in output_from_method]
            for lang in ["de", "fr"]
        }
        self.assertDictEqual(res_comments, expected_comments)

    def test_image_comments(self) -> None:
        expected_comments_of_image = {"en": "Image", "de": "Bild"}
        res_comments_of_image = jsonpath_ng.ext.parse('$[?name="Image"].comments').find(output_from_method)[0].value
        self.assertDictEqual(expected_comments_of_image, res_comments_of_image)

    def test_first_class_properties(self) -> None:
        expected_first_class_properties = [
            ":hasAnthroponym",
            ":isOwnerOf",
            ":correspondsToGenericAnthroponym",
            ":hasAlias",
            ":hasGender",
            ":isDesignatedAs",
            ":hasTitle",
            ":hasStatus",
            ":hasFamilyRelationTo",
            ":hasLifeYearAmount",
            ":hasBirthDate",
            ":hasDeathDate",
            ":hasBibliography",
            ":hasRemarks",
        ]
        res_first_class_properties = [
            match.value for match in jsonpath_ng.parse("$[0].cardinalities[*].propname").find(output_from_method)
        ]
        assert unordered(res_first_class_properties) == expected_first_class_properties

    def test_cardinalities(self) -> None:
        expected_first_class_cardinalities = [
            "1",
            "0-1",
            "0-n",
            "1",
            "0-n",
            "0-1",
            "1-n",
            "0-1",
            "1-n",
            "0-1",
            "0-1",
            "0-1",
            "1-n",
            "1-n",
        ]
        res_first_class_cardinalities = [
            match.value for match in jsonpath_ng.parse("$[0].cardinalities[*].cardinality").find(output_from_method)
        ]
        assert unordered(res_first_class_cardinalities) == expected_first_class_cardinalities


class TestValidateWithSchema:
    # it is not possible to call the method to be tested directly.
    # So let's make a reference to it, so that it can be found by the usage search
    lambda _: e2j._validate_resources([])

    def test_invalid_super(self) -> None:
        expected_msg = regex.escape(
            "\nThe Excel file 'resources.xlsx' did not pass validation.\n"
            "    Section of the problem: 'Resources'\n"
            "    Problematic Resource 'Title'\n"
            "    Located at: Sheet 'classes' | Column 'super' | Row 3\n"
            "    Original Error Message: 'fantasy' is not valid under any of the given schemas"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j.excel2resources("testdata/invalid-testdata/excel2json/resources-invalid-super.xlsx", "")

    def test_sheet_invalid_cardinality(self) -> None:
        expected_msg = regex.escape(
            "\nThe Excel file 'resources.xlsx' did not pass validation.\n"
            "    Section of the problem: 'Resources'\n"
            "    Located at: Sheet 'Owner' | Column 'Cardinality' | Row 3\n"
            "    Original Error Message: '0-2' is not one of ['1', '0-1', '1-n', '0-n']"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j.excel2resources("testdata/invalid-testdata/excel2json/resources-invalid-cardinality.xlsx", "")

    def test_invalid_property(self) -> None:
        expected_msg = regex.escape(
            "\nThe Excel file 'resources.xlsx' did not pass validation.\n"
            "    Section of the problem: 'Resources'\n"
            "    Located at: Sheet 'FamilyMember' | Column 'Property' | Row 7\n"
            "    Original Error Message: ':fan:ta:sy' does not match '^([a-zA-Z_][\\\\w.-]*)?:([\\\\w.-]+)$'"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j.excel2resources("testdata/invalid-testdata/excel2json/resources-invalid-property.xlsx", "")

    def test_duplicate_name(self) -> None:
        expected_msg = regex.escape(
            "The Excel file 'resources.xlsx' contains the following problems:\n\n"
            "The sheet 'classes' has the following problems:\n\n"
            "No duplicates are allowed in the column 'name'\n"
            "The following values appear several times:\n"
            "    - MentionedPerson"
        )
        with pytest.raises(BaseError, match=expected_msg):
            e2j.excel2resources("testdata/invalid-testdata/excel2json/resources-duplicate-name.xlsx", "")

    def test_duplicate_classes_sheet(self) -> None:
        expected_msg = regex.escape(
            "The Excel file 'testdata/invalid-testdata/excel2json/resources-duplicate-classes-sheet.xlsx' "
            "contains the following problems:\n\n"
            "The sheet names inside the same Excel file must be unique. "
            "Using capitalisation or spaces to differentiate sheets is not valid.\n"
            "For example 'sheet' and 'SHEET  ' are considered identical.\n"
            "Under this condition, the following sheet names appear multiple times:\n"
            "    - classes"
        )
        with pytest.raises(InputError, match=expected_msg):
            e2j.excel2resources("testdata/invalid-testdata/excel2json/resources-duplicate-classes-sheet.xlsx", "")


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass


@dataclass
class ProjectDefinition:
    shortcode: str
    shortname: str
    longname: str
    keywords: list[str] | None = None
    descriptions: dict[str, str] | None = None
    groups: list[dict[str, str]] | None = None
    users: list[dict[str, str]] | None = None



import pytest
import regex

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.exceptions import InputError
from dsp_tools.utils.xml_validation import validate_xml_file


def test_validate_xml_data_systematic() -> None:
    assert validate_xml_file(input_file="testdata/xml-data/test-data-systematic.xml")


def test_validate_xml_data_minimal() -> None:
    assert validate_xml_file(input_file="testdata/xml-data/test-data-minimal.xml")


def test_validate_xml_invalid_resource_tag_line_twelve() -> None:
    expected = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 12: Element 'resource', attribute 'invalidtag': The attribute 'invalidtag' is not allowed."
    )
    with pytest.raises(InputError, match=expected):
        validate_xml_file(input_file="testdata/invalid-testdata/xml-data/invalid-resource-tag.xml")


def test_validate_xml_invalid_resource_tag_problem() -> None:
    expected_msg = regex.escape(
        "Angular brackets in the format of <text> were found in text properties with encoding=utf8.\n"
        "Please note that these will not be recognised as formatting in the text field, "
        "but will be displayed as-is.\n"
        "The following resources of your XML file contain angular brackets:\n"
        "    - line 13: resource 'the_only_resource', property ':test'\n"
        "    - line 14: resource 'the_only_resource', property ':test'\n"
        "    - line 15: resource 'the_only_resource', property ':test'\n"
        "    - line 16: resource 'the_only_resource', property ':test'",
    )
    with pytest.warns(DspToolsUserWarning, match=expected_msg):
        validate_xml_file(input_file="testdata/invalid-testdata/xml-data/utf8-text-with-xml-tags.xml")


def test_validate_xml_data_duplicate_iri() -> None:
    expected_msg = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 19: Element 'resource': Duplicate key-sequence ['http://rdfh.ch/4123/54SYvWF0QUW6a'] "
        "in unique identity-constraint 'IRI_attribute_of_resource_must_be_unique'."
    )
    with pytest.raises(InputError, match=expected_msg):
        validate_xml_file(input_file="testdata/invalid-testdata/xml-data/duplicate-iri.xml")


def test_validate_xml_duplicate_ark() -> None:
    expected_msg = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 19: Element 'resource': Duplicate key-sequence ['ark:/72163/4123-31ec6eab334-a.2022829'] "
        "in unique identity-constraint 'ARK_attribute_of_resource_must_be_unique'."
    )
    with pytest.raises(InputError, match=expected_msg):
        validate_xml_file(input_file="testdata/invalid-testdata/xml-data/duplicate-ark.xml")


def test_validate_xml_empty_label() -> None:
    expected_msg = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 11: Element 'resource', attribute 'label': [facet 'minLength'] "
        "The value '' has a length of '0'; this underruns the allowed minimum length of '1'."
    )
    with pytest.raises(InputError, match=expected_msg):
        validate_xml_file(input_file="testdata/invalid-testdata/xml-data/empty-label.xml")


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
from lxml import etree

from dsp_tools.commands.xml_validate.deserialise_input import _deserialise_all_resources
from dsp_tools.commands.xml_validate.deserialise_input import _deserialise_one_property
from dsp_tools.commands.xml_validate.deserialise_input import _deserialise_one_resource
from dsp_tools.commands.xml_validate.models.data_deserialised import BooleanValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ColorValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import DateValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import DecimalValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import GeonameValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import IntValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import LinkValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ListValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import RegionDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ResourceDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import RichtextDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import SimpleTextDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import TimeValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import UriValueDeserialised


class TestResource:
    def test_empty(self, resource_empty: etree._Element) -> None:
        res = _deserialise_one_resource(resource_empty)
        assert isinstance(res, ResourceDeserialised)
        assert res.res_id == "one"
        assert res.res_class == "http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything"
        assert res.label == "lbl"
        assert len(res.values) == 0

    def test_with_props(self, root_resource_with_props: etree._Element) -> None:
        res_list = _deserialise_all_resources(root_resource_with_props).resources
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, ResourceDeserialised)
        assert res.res_id == "one"
        assert res.res_class == "http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything"
        assert res.label == "lbl"
        assert len(res.values) == 3

    def test_region(self, root_resource_region: etree._Element) -> None:
        res_list = _deserialise_all_resources(root_resource_region).resources
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, RegionDeserialised)
        assert res.res_id == "region_1"


class TestBooleanValue:
    def test_corr(self, boolean_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(boolean_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, BooleanValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean"
        assert res.object_value == "true"


class TestColorValue:
    def test_corr(self, color_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(color_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, ColorValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testColor"
        assert res.object_value == "#00ff00"

    def test_several(self, color_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(color_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, ColorValueDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testColor"
        assert res[0].object_value == "#00ff00"
        assert res[1].object_value == "#00ff11"


class TestDateValue:
    def test_corr(self, date_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(date_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, DateValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1"
        assert res.object_value == "JULIAN:BCE:0700:BCE:0600"

    def test_several(self, date_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(date_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, DateValueDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1"
        assert res[0].object_value == "JULIAN:BCE:0700:BCE:0600"
        assert res[1].object_value == "ISLAMIC:BCE:0700:BCE:0600"


class TestDecimalValue:
    def test_corr(self, decimal_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(decimal_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, DecimalValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText"
        assert res.object_value == "2.71"

    def test_several(self, decimal_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(decimal_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, DecimalValueDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText"
        assert res[0].object_value == "1.0"
        assert res[1].object_value == "2.0"


class TestGeonameValue:
    def test_corr(self, geoname_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(geoname_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, GeonameValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname"
        assert res.object_value == "1111111"

    def test_several(self, geoname_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(geoname_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, GeonameValueDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname"
        assert res[0].object_value == "1111111"
        assert res[1].object_value == "2222222"


class TestIntValue:
    def test_corr(self, integer_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(integer_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, IntValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText"
        assert res.object_value == "1"

    def test_several(self, integer_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(integer_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, IntValueDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText"
        assert res[0].object_value == "1"
        assert res[1].object_value == "2"


class TestListValue:
    def test_corr(self, list_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(list_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, ListValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp"
        assert res.list_name == "onlyList"
        assert res.object_value == "n1"

    def test_several(self, list_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(list_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, ListValueDeserialised) for x in res])
        one = res[0]
        assert isinstance(one, ListValueDeserialised)
        assert one.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp"
        assert one.list_name == "onlyList"
        assert res[0].object_value == "n1"
        assert res[1].object_value == "n2"


class TestSimpleTextValue:
    def test_corr(self, text_simpletext_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(text_simpletext_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, SimpleTextDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea"
        assert res.object_value == "Text"

    def test_several(self, text_simpletext_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(text_simpletext_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, SimpleTextDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testSimpleText"
        assert res[0].object_value == "Text 1"
        assert res[1].object_value == "Text 2"

    def test_wrong(self, text_simpletext_value_wrong: etree._Element) -> None:
        res_list = _deserialise_one_property(text_simpletext_value_wrong)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, SimpleTextDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testSimpleText"
        assert not res.object_value


class TestRichtextValue:
    def test_corr(self, text_richtext_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(text_richtext_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, RichtextDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext"
        assert res.object_value == "Text"

    def test_several(self, text_richtext_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(text_richtext_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, RichtextDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext"
        assert res[0].object_value == "Text 1"
        assert res[1].object_value == "Text 2"


class TestTimeValue:
    def test_corr(self, time_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(time_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, TimeValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue"
        assert res.object_value == "2019-10-23T13:45:12.01-14:00"

    def test_several(self, time_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(time_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, TimeValueDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue"
        assert res[0].object_value == "2019-10-23T13:45:12.01-14:00"
        assert res[1].object_value == "2019-10-23T13:45:12.01-08:00"


class TestUriValue:
    def test_corr(self, uri_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(uri_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, UriValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue"
        assert res.object_value == "https://dasch.swiss"

    def test_several(self, uri_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(uri_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, UriValueDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue"
        assert res[0].object_value == "https://dasch.swiss"
        assert res[1].object_value == "https://app.dasch.swiss"


class TestLinkValue:
    def test_corr(self, resptr_value_corr: etree._Element) -> None:
        res_list = _deserialise_one_property(resptr_value_corr)
        assert len(res_list) == 1
        res = res_list[0]
        assert isinstance(res, LinkValueDeserialised)
        assert res.prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo"
        assert res.object_value == "id_1"

    def test_several(self, resptr_value_corr_several: etree._Element) -> None:
        res = _deserialise_one_property(resptr_value_corr_several)
        assert len(res) == 2
        assert all([isinstance(x, LinkValueDeserialised) for x in res])
        assert res[0].prop_name == "http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo"
        assert res[0].object_value == "id_1"
        assert res[1].object_value == "id_2"


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from dsp_tools.commands.xml_validate.models.data_deserialised import BooleanValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ColorValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import DateValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import DecimalValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import GeonameValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import IntValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import LinkValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ListValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ResourceDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import RichtextDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import SimpleTextDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import TimeValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import UriValueDeserialised


@pytest.fixture
def resource_deserialised_with_values(
    boolean_value_deserialised_corr: BooleanValueDeserialised,
) -> ResourceDeserialised:
    return ResourceDeserialised(
        res_id="id",
        res_class="http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything",
        label="lbl",
        values=[boolean_value_deserialised_corr],
    )


@pytest.fixture
def resource_deserialised_no_values() -> ResourceDeserialised:
    return ResourceDeserialised(
        res_id="id", res_class="http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything", label="lbl", values=[]
    )


@pytest.fixture
def boolean_value_deserialised_corr() -> BooleanValueDeserialised:
    return BooleanValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean", "false")


@pytest.fixture
def boolean_value_deserialised_zero() -> BooleanValueDeserialised:
    return BooleanValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean", "0")


@pytest.fixture
def boolean_value_deserialised_one() -> BooleanValueDeserialised:
    return BooleanValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean", "1")


@pytest.fixture
def boolean_value_deserialised_none() -> BooleanValueDeserialised:
    return BooleanValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean", None)


@pytest.fixture
def color_value_deserialised_corr() -> ColorValueDeserialised:
    return ColorValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testColor", "#00ff00")


@pytest.fixture
def color_value_deserialised_none() -> ColorValueDeserialised:
    return ColorValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testColor", None)


@pytest.fixture
def date_value_deserialised_corr() -> DateValueDeserialised:
    return DateValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1", "JULIAN:BCE:0700:BCE:0600")


@pytest.fixture
def date_value_deserialised_none() -> DateValueDeserialised:
    return DateValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1", None)


@pytest.fixture
def decimal_value_deserialised_corr() -> DecimalValueDeserialised:
    return DecimalValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText", "1.2")


@pytest.fixture
def decimal_value_deserialised_none() -> DecimalValueDeserialised:
    return DecimalValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText", None)


@pytest.fixture
def geoname_value_deserialised_corr() -> GeonameValueDeserialised:
    return GeonameValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname", "1241345")


@pytest.fixture
def geoname_value_deserialised_none() -> GeonameValueDeserialised:
    return GeonameValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname", None)


@pytest.fixture
def int_value_deserialised_corr() -> IntValueDeserialised:
    return IntValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText", "1")


@pytest.fixture
def int_value_deserialised_none() -> IntValueDeserialised:
    return IntValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText", None)


@pytest.fixture
def link_value_deserialised_corr() -> LinkValueDeserialised:
    return LinkValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo", "link-id")


@pytest.fixture
def link_value_deserialised_none() -> LinkValueDeserialised:
    return LinkValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo", None)


@pytest.fixture
def list_value_deserialised_corr() -> ListValueDeserialised:
    return ListValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp", "n1", "onlyList")


@pytest.fixture
def list_value_deserialised_none() -> ListValueDeserialised:
    return ListValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp", None, "onlyList")


@pytest.fixture
def simple_text_deserialised_corr() -> SimpleTextDeserialised:
    return SimpleTextDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea", "simple text")


@pytest.fixture
def simple_text_deserialised_none() -> SimpleTextDeserialised:
    return SimpleTextDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea", None)


@pytest.fixture
def richtext_deserialised_corr() -> RichtextDeserialised:
    return RichtextDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext", "rich text")


@pytest.fixture
def richtext_deserialised_none() -> RichtextDeserialised:
    return RichtextDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext", None)


@pytest.fixture
def time_value_deserialised_corr() -> TimeValueDeserialised:
    return TimeValueDeserialised(
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue", "2019-10-23T13:45:12.01-14:00"
    )


@pytest.fixture
def time_value_deserialised_none() -> TimeValueDeserialised:
    return TimeValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue", None)


@pytest.fixture
def uri_value_deserialised_corr() -> UriValueDeserialised:
    return UriValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue", "https://dasch.swiss")


@pytest.fixture
def uri_value_deserialised_none() -> UriValueDeserialised:
    return UriValueDeserialised("http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue", None)



from __future__ import annotations

from pathlib import Path

import regex
from lxml import etree

from dsp_tools.models.exceptions import UserError
from dsp_tools.utils.iri_util import is_resource_iri
from dsp_tools.utils.xml_utils import parse_xml_file
from dsp_tools.utils.xml_utils import remove_comments_from_element_tree
from dsp_tools.utils.xml_utils import transform_special_tags_make_localname
from dsp_tools.utils.xml_validation import validate_xml


def validate_and_parse(input_file: Path) -> tuple[etree._Element, str, str]:
    """Parse and validate an XML file.

    Args:
        input_file: Path to the XML file

    Returns:
        The root element of the parsed XML file, the shortcode, and the default ontology
    """
    root = parse_xml_file(input_file)
    root = remove_comments_from_element_tree(root)

    validate_xml(root)
    root = transform_special_tags_make_localname(root)
    _check_if_link_targets_exist(root)
    shortcode = root.attrib["shortcode"]
    default_ontology = root.attrib["default-ontology"]
    return root, shortcode, default_ontology


def _check_if_link_targets_exist(root: etree._Element) -> None:
    """
    Make sure that all targets of links (resptr and salsah-links)
    are either IRIs or IDs that exist in the present XML file.

    Args:
        root: parsed XML file

    Raises:
        UserError: if a link target does not exist in the XML file
    """
    resptr_errors = _check_if_resptr_targets_exist(root)
    salsah_errors = _check_if_salsah_targets_exist(root)
    if errors := resptr_errors + salsah_errors:
        sep = "\n - "
        msg = f"It is not possible to upload the XML file, because it contains invalid links:{sep}{sep.join(errors)}"
        raise UserError(msg)


def _check_if_resptr_targets_exist(root: etree._Element) -> list[str]:
    link_values = [x for x in root.iter() if x.tag == "resptr"]
    resource_ids = [x.attrib["id"] for x in root.iter() if x.tag == "resource"]
    invalid_link_values = [x for x in link_values if x.text not in resource_ids]
    invalid_link_values = [x for x in invalid_link_values if not is_resource_iri(str(x.text))]
    errors = []
    for inv in invalid_link_values:
        prop_name = next(inv.iterancestors(tag="resptr-prop")).attrib["name"]
        res_id = next(inv.iterancestors(tag="resource")).attrib["id"]
        errors.append(f"Resource '{res_id}', property '{prop_name}' has an invalid link target '{inv.text}'")
    return errors


def _check_if_salsah_targets_exist(root: etree._Element) -> list[str]:
    link_values = [x for x in root.iter() if x.tag == "a"]
    resource_ids = [x.attrib["id"] for x in root.iter() if x.tag == "resource"]
    invalid_link_values = [x for x in link_values if regex.sub(r"IRI:|:IRI", "", x.attrib["href"]) not in resource_ids]
    invalid_link_values = [x for x in invalid_link_values if x.attrib.get("class") == "salsah-link"]
    invalid_link_values = [x for x in invalid_link_values if not is_resource_iri(x.attrib["href"])]
    errors = []
    for inv in invalid_link_values:
        prop_name = next(inv.iterancestors(tag="text-prop")).attrib["name"]
        res_id = next(inv.iterancestors(tag="resource")).attrib["id"]
        errors.append(f"Resource '{res_id}', property '{prop_name}' has an invalid link target '{inv.attrib['href']}'")
    return errors


def check_if_bitstreams_exist(root: etree._Element, imgdir: str) -> None:
    """
    Make sure that all bitstreams referenced in the XML file exist in the imgdir.

    Args:
        root: parsed XML file
        imgdir: folder where the bitstreams are stored

    Raises:
        UserError: if a bitstream does not exist in the imgdir
    """
    multimedia_resources = [x for x in root if any((y.tag == "bitstream" for y in x.iter()))]
    for res in multimedia_resources:
        pth = next(Path(x.text) for x in res.iter() if x.tag == "bitstream" and x.text)
        if not Path(imgdir / pth).is_file():
            raise UserError(
                f"Bitstream '{pth!s}' of resource '{res.attrib['label']}' does not exist in the imgdir '{imgdir}'."
            )
